var cluster                 =   require('cluster'),
    domain                  =   require('domain'),
    numCPUs                 =   require('os').cpus().length,
    
    _                       =   require('underscore'),
    express                 =   require('express'),
    session                 =   require('express-session'),
    redis                   =   require('connect-redis')(session),
    bodyParser              =   require('body-parser'),
    cookieParser            =   require('cookie-parser'),
    compression             =   require('compression'),
    limits                  =   require('limits'),
    vhost                   =   require('vhost'),
    
    auth                    =   require('./auth/index'),
    crypto                  =   require('./crypto/index'),
    database                =   require('./database/index'),
    error                   =   require('./error/index'),
    interfaces              =   require('./interfaces/index'),
    logger                  =   require('./logger/index'),
    mailer                  =   require('./mailer/index'),
    notify                  =   require('./notify/index'),
    parser                  =   require('./parser/index'),
    search                  =   require('./search/index'),
    storage                 =   require('./storage/index'),
    utilities               =   require('./utilities/index'),
    validator               =   require('./validator/index'),
    
    config                  =   { },
    servers                 =   [],
    killTimers              =   [];



/******************** Middleware ********************/

function bindToDomain (req, res, next) {
    var reqDomain = domain.create();
    reqDomain.add(req);
    reqDomain.add(res);
    
    reqDomain.on('error', function(error) {
        utilities.logger.error ( '|||****************** Unhandled Error ******************|||' );
        utilities.logger.error ( 'Cluster ID: ', cluster.worker.id );
        utilities.logger.error ( error.stack );
        utilities.logger.error ( '|||*****************************************************|||' );
        
        // Note: we're in dangerous territory!
        // By definition, something unexpected occurred,
        // which we probably didn't want.
        // Anything can happen now!  Be very careful!
        try {
            
            if (numCPUs < 2) {
                servers[cluster.worker.id].close();
                cluster.worker.disconnect();
                process.exit(1);
            } else {
                // make sure we close down within 30 seconds
                killTimers[cluster.worker.id] = setTimeout(function() { process.exit(1); }, 30000);
                // But don't keep the process open just for that!
                killTimers[cluster.worker.id].unref();
                // stop taking new requests.
                servers[cluster.worker.id].close();
                // Let the master know we're dead.  This will trigger a
                // 'disconnect' in the cluster master, and then it will fork
                // a new worker.
                cluster.worker.disconnect();
            }

            // try to send an error to the request that triggered the problem
            //res.statusCode = 500;
            res.setHeader('content-type', 'application/json');
            res.end(JSON.stringify({ error: 200 }));

        } catch (sendError) {
            // oh well, not much we can do at this point.
            utilities.logger.error ('Error sending 500!');
            utilities.logger.error(sendError.stack);
        }
    });

    reqDomain.run(next);
}

function allowCORS (req, res, next) {
    if (req.headers.origin) {
        utilities.logger.info  ('CORS Origin: ' + req.headers.origin );
        utilities.logger.info  ('CORS Headers: ' + JSON.stringify(req.headers) );
        var corsType = config.server.cors[req.headers.origin] || config.server.cors.all;
        utilities.logger.info  ('CORS Type: ' + JSON.stringify (corsType) );
        if ( corsType.enabled ) {
            utilities.logger.info ('CORS hostname: ' + req.hostname);
            utilities.logger.info ('Requester IP: ' + req.ip);

            res.header ('Access-Control-Allow-Origin',      req.headers.origin );
            
            if (corsType.methods)       res.header ('Access-Control-Allow-Methods',     corsType.methods );
            if (corsType.headers)       res.header ('Access-Control-Allow-Headers',     corsType.headers );
            if (corsType.credentials)   res.header ('Access-Control-Allow-Credentials', true );
            
            if (req.method == 'OPTIONS') return res.sendStatus(200);
        }
    }
    next();
}

function redirectToHttps (req, res, next) {
    if (!req.secure) res.redirect('https://' + req.headers.host + req.url);
    next();
}

function logRequests (req, res, next) {
    if (req.url !== config.server.healthCheckPath) utilities.logger.info(req.method + " " + req.url); 
    next();
}

function healthCheck (req, res, next) {
    if (req.url === config.server.healthCheckPath) res.sendStatus(200); 
    else next();
}
/****************************************************/


function start(server) {
    var http =  config.server.ssl.enabled ? require('https') : require('http');
    var serverInstance = config.server.ssl.enabled ? http.createServer(config.server.ssl.options, server) : http.createServer(server); 

    //utilities.logger.info(" -=** Server started on port " + ( config.server.port ) + " **=-");
    
    // Websockets
    if (config.interfaces.websockets.enabled) interfaces.initWebsockets ( serverInstance, config.server );

    return serverInstance.listen ( config.server.port );
}

function configure (server) {
	var hosts = [];

    // create domains for each req, res pair.
    server.use (bindToDomain);

    server.enable ("trust proxy");

	if ( config.server.env.prod ) server.use ( compression() );
    else server.use (logRequests);

    server.use (healthCheck);

    // Generic Error Logging
    //server.use( function(err, req, res, next) { if (err) {  utilities.logger.error(err.stack); res.sendStatus(err.status); } else next(); });

    server.use ( cookieParser(config.server.session.secret) );
    server.use ( bodyParser.json());
    server.use ( bodyParser.urlencoded ( { extended: true } ) );
    server.use ( session (config.server.session) );
    
    if ( config.server.jsonp ) server.enable ("jsonp callback");
    if ( config.server.cors.enabled ) server.use ( allowCORS );
    
    server.use ( limits (config.server.limits) );

    //server.use (redirectToHttps);

    // set up authentication
    auth.init(server, config.auth);

    // Set up each webserver host
    if (config.server.staticHosting.enabled) {
        _.each (config.server.staticHosting.hosts, function(host, name){
            hosts[name] = express();
            hosts[name].use ( express.static( host.localDir, { maxAge: config.times.oneDay }));
            server.use ( vhost (host.url, hosts[name]) );
        });
    }
    
    //REST API
    if (config.interfaces.rest.enabled) interfaces.initREST ( server, config.server );

}

//// Public --------------------------------------------------------------------------------------


exports.init = function (options) {
        
    if (options && _.isObject(options)) _.extend(config, options);


    config.server.session.store = new redis (config.server.redisStore);

    error.init                  ( config.error );
    crypto.init                 ( config.crypto );
    validator.init              ( config.validator );
    interfaces.setControllers   ( config.interfaces.controllers );
    notify.init                 ( config.notify );
    search.init                 ( config.search );
    storage.init                ( config.storage );
    mailer.init                 ( config.mailer );

    database.init ( config.database, function (e) {
        if (e) {
            setTimeout ( databaseInit, 5000);
        } else { 
            servers [cluster.worker.id] = express();
            configure (servers[cluster.worker.id]);
            servers [cluster.worker.id] = start (servers[cluster.worker.id]);
        }
    });
    
};


exports.auth =         auth;
exports.crypto =       crypto;
exports.database =     database;
exports.error =        error;
exports.interfaces =   interfaces;
exports.logger =       logger;
exports.mailer =       mailer;
exports.notify =       notify;
exports.parser =       parser;
exports.search =       search;
exports.storage =      storage;
exports.utilities =    utilities;
exports.validator =    validator;